@@ -1,40 +1,39 @@
 """A base class for contents managers."""
-
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
-
-from fnmatch import fnmatch
 import itertools
 import json
-import os
 import re
+import warnings
+from fnmatch import fnmatch
 
-from tornado.web import HTTPError, RequestHandler
+from ipython_genutils.importstring import import_item
+from nbformat import sign
+from nbformat import validate as validate_nb
+from nbformat import ValidationError
+from nbformat.v4 import new_notebook
+from tornado.web import HTTPError
+from tornado.web import RequestHandler
+from traitlets import Any
+from traitlets import Bool
+from traitlets import default
+from traitlets import Dict
+from traitlets import Instance
+from traitlets import List
+from traitlets import TraitError
+from traitlets import Type
+from traitlets import Unicode
+from traitlets import validate
+from traitlets.config.configurable import LoggingConfigurable
 
 from ...files.handlers import FilesHandler
-from .checkpoints import Checkpoints, AsyncCheckpoints
-from traitlets.config.configurable import LoggingConfigurable
-from nbformat import sign, validate as validate_nb, ValidationError
-from nbformat.v4 import new_notebook
-from ipython_genutils.importstring import import_item
-from traitlets import (
-    Any,
-    Bool,
-    Dict,
-    Instance,
-    List,
-    TraitError,
-    Type,
-    Unicode,
-    validate,
-    default,
-)
-from ipython_genutils.py3compat import string_types
+from .checkpoints import AsyncCheckpoints
+from .checkpoints import Checkpoints
 from jupyter_server.transutils import _i18n
 from jupyter_server.utils import ensure_async
 
 
-copy_pat = re.compile(r'\-Copy\d*\.')
+copy_pat = re.compile(r"\-Copy\d*\.")
 
 
 class ContentsManager(LoggingConfigurable):
@@ -56,34 +55,52 @@ class ContentsManager(LoggingConfigurable):
 
     """
 
-    root_dir = Unicode('/', config=True)
+    root_dir = Unicode("/", config=True)
 
     allow_hidden = Bool(False, config=True, help="Allow access to hidden files")
 
     notary = Instance(sign.NotebookNotary)
+
     def _notary_default(self):
         return sign.NotebookNotary(parent=self)
 
-    hide_globs = List(Unicode(), [
-            u'__pycache__', '*.pyc', '*.pyo',
-            '.DS_Store', '*.so', '*.dylib', '*~',
-        ], config=True, help="""
+    hide_globs = List(
+        Unicode(),
+        [
+            "__pycache__",
+            "*.pyc",
+            "*.pyo",
+            ".DS_Store",
+            "*.so",
+            "*.dylib",
+            "*~",
+        ],
+        config=True,
+        help="""
         Glob patterns to hide in file and directory listings.
-    """)
-
-    untitled_notebook = Unicode(_i18n("Untitled"), config=True,
-        help="The base name used when creating untitled notebooks."
+    """,
     )
 
-    untitled_file = Unicode("untitled", config=True,
-        help="The base name used when creating untitled files."
+    untitled_notebook = Unicode(
+        _i18n("Untitled"),
+        config=True,
+        help="The base name used when creating untitled notebooks.",
     )
 
-    untitled_directory = Unicode("Untitled Folder", config=True,
-        help="The base name used when creating untitled directories."
+    untitled_file = Unicode(
+        "untitled", config=True, help="The base name used when creating untitled files."
     )
 
-    pre_save_hook = Any(None, config=True, allow_none=True,
+    untitled_directory = Unicode(
+        "Untitled Folder",
+        config=True,
+        help="The base name used when creating untitled directories.",
+    )
+
+    pre_save_hook = Any(
+        None,
+        config=True,
+        allow_none=True,
         help="""Python callable or importstring thereof
 
         To be called on a contents model prior to save.
@@ -100,20 +117,65 @@ class ContentsManager(LoggingConfigurable):
           Modifying this dict will affect the file that is stored.
         - path: the API path of the save destination
         - contents_manager: this ContentsManager instance
-        """
+        """,
     )
 
-    @validate('pre_save_hook')
+    @validate("pre_save_hook")
     def _validate_pre_save_hook(self, proposal):
-        value = proposal['value']
-        if isinstance(value, string_types):
+        value = proposal["value"]
+        if isinstance(value, str):
             value = import_item(self.pre_save_hook)
         if not callable(value):
             raise TraitError("pre_save_hook must be callable")
+        if self.pre_save_hook is not None:
+            warnings.warn(
+                f"Overriding existing pre_save_hook ({self.pre_save_hook.__name__}) with a new one ({value.__name__}).",
+                stacklevel=2,
+            )
+        return value
+
+    post_save_hook = Any(
+        None,
+        config=True,
+        allow_none=True,
+        help="""Python callable or importstring thereof
+
+        to be called on the path of a file just saved.
+
+        This can be used to process the file on disk,
+        such as converting the notebook to a script or HTML via nbconvert.
+
+        It will be called as (all arguments passed by keyword)::
+
+            hook(os_path=os_path, model=model, contents_manager=instance)
+
+        - path: the filesystem path to the file just written
+        - model: the model representing the file
+        - contents_manager: this ContentsManager instance
+        """,
+    )
+
+    @validate("post_save_hook")
+    def _validate_post_save_hook(self, proposal):
+        value = proposal["value"]
+        if isinstance(value, str):
+            value = import_item(value)
+        if not callable(value):
+            raise TraitError("post_save_hook must be callable")
+        if self.post_save_hook is not None:
+            warnings.warn(
+                f"Overriding existing post_save_hook ({self.post_save_hook.__name__}) with a new one ({value.__name__}).",
+                stacklevel=2,
+            )
         return value
 
     def run_pre_save_hook(self, model, path, **kwargs):
         """Run the pre-save hook if defined, and log errors"""
+        warnings.warn(
+            "run_pre_save_hook is deprecated, use run_pre_save_hooks instead.",
+            DeprecationWarning,
+            stacklevel=2,
+        )
         if self.pre_save_hook:
             try:
                 self.log.debug("Running pre-save hook on %s", path)
@@ -127,15 +189,86 @@ class ContentsManager(LoggingConfigurable):
                 # which could cause frustrating data loss
                 self.log.error("Pre-save hook failed on %s", path, exc_info=True)
 
+    def run_post_save_hook(self, model, os_path):
+        """Run the post-save hook if defined, and log errors"""
+        warnings.warn(
+            "run_post_save_hook is deprecated, use run_post_save_hooks instead.",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        if self.post_save_hook:
+            try:
+                self.log.debug("Running post-save hook on %s", os_path)
+                self.post_save_hook(os_path=os_path, model=model, contents_manager=self)
+            except Exception as e:
+                self.log.error("Post-save hook failed o-n %s", os_path, exc_info=True)
+                raise HTTPError(500, "Unexpected error while running post hook save: %s" % e) from e
+
+    _pre_save_hooks = List()
+    _post_save_hooks = List()
+
+    def register_pre_save_hook(self, hook):
+        if isinstance(hook, str):
+            hook = import_item(hook)
+        if not callable(hook):
+            raise RuntimeError("hook must be callable")
+        self._pre_save_hooks.append(hook)
+
+    def register_post_save_hook(self, hook):
+        if isinstance(hook, str):
+            hook = import_item(hook)
+        if not callable(hook):
+            raise RuntimeError("hook must be callable")
+        self._post_save_hooks.append(hook)
+
+    def run_pre_save_hooks(self, model, path, **kwargs):
+        """Run the pre-save hooks if any, and log errors"""
+        pre_save_hooks = [self.pre_save_hook] if self.pre_save_hook is not None else []
+        pre_save_hooks += self._pre_save_hooks
+        for pre_save_hook in pre_save_hooks:
+            try:
+                self.log.debug("Running pre-save hook on %s", path)
+                pre_save_hook(model=model, path=path, contents_manager=self, **kwargs)
+            except HTTPError:
+                # allow custom HTTPErrors to raise,
+                # rejecting the save with a message.
+                raise
+            except Exception:
+                # unhandled errors don't prevent saving,
+                # which could cause frustrating data loss
+                self.log.error(
+                    "Pre-save hook %s failed on %s",
+                    pre_save_hook.__name__,
+                    path,
+                    exc_info=True,
+                )
+
+    def run_post_save_hooks(self, model, os_path):
+        """Run the post-save hooks if any, and log errors"""
+        post_save_hooks = [self.post_save_hook] if self.post_save_hook is not None else []
+        post_save_hooks += self._post_save_hooks
+        for post_save_hook in post_save_hooks:
+            try:
+                self.log.debug("Running post-save hook on %s", os_path)
+                post_save_hook(os_path=os_path, model=model, contents_manager=self)
+            except Exception as e:
+                self.log.error(
+                    "Post-save %s hook failed on %s",
+                    post_save_hook.__name__,
+                    os_path,
+                    exc_info=True,
+                )
+                raise HTTPError(500, "Unexpected error while running post hook save: %s" % e) from e
+
     checkpoints_class = Type(Checkpoints, config=True)
     checkpoints = Instance(Checkpoints, config=True)
     checkpoints_kwargs = Dict(config=True)
 
-    @default('checkpoints')
+    @default("checkpoints")
     def _default_checkpoints(self):
         return self.checkpoints_class(**self.checkpoints_kwargs)
 
-    @default('checkpoints_kwargs')
+    @default("checkpoints_kwargs")
     def _default_checkpoints_kwargs(self):
         return dict(
             parent=self,
@@ -143,7 +276,10 @@ class ContentsManager(LoggingConfigurable):
         )
 
     files_handler_class = Type(
-        FilesHandler, klass=RequestHandler, allow_none=True, config=True,
+        FilesHandler,
+        klass=RequestHandler,
+        allow_none=True,
+        config=True,
         help="""handler class to use when serving raw file requests.
 
         Default is a fallback that talks to the ContentsManager API,
@@ -153,7 +289,7 @@ class ContentsManager(LoggingConfigurable):
         which will be much more efficient.
 
         Access to these files should be Authenticated.
-        """
+        """,
     )
 
     files_handler_params = Dict(
@@ -162,7 +298,7 @@ class ContentsManager(LoggingConfigurable):
 
         For example, StaticFileHandlers generally expect a `path` argument
         specifying the root directory from which to serve files.
-        """
+        """,
     )
 
     def get_extra_handlers(self):
@@ -172,9 +308,7 @@ class ContentsManager(LoggingConfigurable):
         """
         handlers = []
         if self.files_handler_class:
-            handlers.append(
-                (r"/files/(.*)", self.files_handler_class, self.files_handler_params)
-            )
+            handlers.append((r"/files/(.*)", self.files_handler_class, self.files_handler_params))
         return handlers
 
     # ContentsManager API part 1: methods that must be
@@ -216,7 +350,7 @@ class ContentsManager(LoggingConfigurable):
         """
         raise NotImplementedError
 
-    def file_exists(self, path=''):
+    def file_exists(self, path=""):
         """Does a file exist at the given path?
 
         Like os.path.isfile
@@ -233,7 +367,7 @@ class ContentsManager(LoggingConfigurable):
         exists : bool
             Whether the file exists.
         """
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     def exists(self, path):
         """Does a file or directory exist at the given path?
@@ -254,7 +388,7 @@ class ContentsManager(LoggingConfigurable):
 
     def get(self, path, content=True, type=None, format=None):
         """Get a file or directory model."""
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     def save(self, model, path):
         """
@@ -264,22 +398,22 @@ class ContentsManager(LoggingConfigurable):
         should call self.run_pre_save_hook(model=model, path=path) prior to
         writing any data.
         """
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     def delete_file(self, path):
         """Delete the file or directory at path."""
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     def rename_file(self, old_path, new_path):
         """Rename a file or directory."""
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     # ContentsManager API part 2: methods that have useable default
     # implementations, but can be overridden in subclasses.
 
     def delete(self, path):
         """Delete a file/directory and any associated checkpoints."""
-        path = path.strip('/')
+        path = path.strip("/")
         if not path:
             raise HTTPError(400, "Can't delete root")
         self.delete_file(path)
@@ -296,8 +430,8 @@ class ContentsManager(LoggingConfigurable):
         For use in PATCH requests, to enable renaming a file without
         re-uploading its contents. Only used for renaming at the moment.
         """
-        path = path.strip('/')
-        new_path = model.get('path', path).strip('/')
+        path = path.strip("/")
+        new_path = model.get("path", path).strip("/")
         if path != new_path:
             self.rename(path, new_path)
         model = self.get(new_path, content=False)
@@ -316,9 +450,9 @@ class ContentsManager(LoggingConfigurable):
         notebook server. FileContentsManager overrides this to use the
         directory containing the notebook.
         """
-        return ''
+        return ""
 
-    def increment_filename(self, filename, path='', insert=''):
+    def increment_filename(self, filename, path="", insert=""):
         """Increment a filename until it is unique.
 
         Parameters
@@ -327,7 +461,7 @@ class ContentsManager(LoggingConfigurable):
             The name of a file, including extension
         path : unicode
             The API path of the target's directory
-        insert: unicode
+        insert : unicode
             The characters to insert after the base filename
 
         Returns
@@ -336,35 +470,47 @@ class ContentsManager(LoggingConfigurable):
             A filename that is unique, based on the input filename.
         """
         # Extract the full suffix from the filename (e.g. .tar.gz)
-        path = path.strip('/')
-        basename, dot, ext = filename.rpartition('.')
-        if ext != 'ipynb':
-                basename, dot, ext = filename.partition('.')
+        path = path.strip("/")
+        basename, dot, ext = filename.rpartition(".")
+        if ext != "ipynb":
+            basename, dot, ext = filename.partition(".")
 
         suffix = dot + ext
 
         for i in itertools.count():
             if i:
-                insert_i = '{}{}'.format(insert, i)
+                insert_i = "{}{}".format(insert, i)
             else:
-                insert_i = ''
-            name = u'{basename}{insert}{suffix}'.format(basename=basename,
-                insert=insert_i, suffix=suffix)
-            if not self.exists(u'{}/{}'.format(path, name)):
+                insert_i = ""
+            name = "{basename}{insert}{suffix}".format(
+                basename=basename, insert=insert_i, suffix=suffix
+            )
+            if not self.exists("{}/{}".format(path, name)):
                 break
         return name
 
-    def validate_notebook_model(self, model):
+    def validate_notebook_model(self, model, validation_error=None):
         """Add failed-validation message to model"""
         try:
-            validate_nb(model['content'])
+            # If we're given a validation_error dictionary, extract the exception
+            # from it and raise the exception, else call nbformat's validate method
+            # to determine if the notebook is valid.  This 'else' condition may
+            # pertain to server extension not using the server's notebook read/write
+            # functions.
+            if validation_error is not None:
+                e = validation_error.get("ValidationError")
+                if isinstance(e, ValidationError):
+                    raise e
+            else:
+                validate_nb(model["content"])
         except ValidationError as e:
-            model['message'] = u'Notebook validation failed: {}:\n{}'.format(
-                e.message, json.dumps(e.instance, indent=1, default=lambda obj: '<UNKNOWN>'),
+            model["message"] = "Notebook validation failed: {}:\n{}".format(
+                e.message,
+                json.dumps(e.instance, indent=1, default=lambda obj: "<UNKNOWN>"),
             )
         return model
 
-    def new_untitled(self, path='', type='', ext=''):
+    def new_untitled(self, path="", type="", ext=""):
         """Create a new untitled file or directory in path
 
         path must be a directory
@@ -373,58 +519,58 @@ class ContentsManager(LoggingConfigurable):
 
         Use `new` to create files with a fully specified path (including filename).
         """
-        path = path.strip('/')
+        path = path.strip("/")
         if not self.dir_exists(path):
-            raise HTTPError(404, 'No such directory: %s' % path)
+            raise HTTPError(404, "No such directory: %s" % path)
 
         model = {}
         if type:
-            model['type'] = type
+            model["type"] = type
 
-        if ext == '.ipynb':
-            model.setdefault('type', 'notebook')
+        if ext == ".ipynb":
+            model.setdefault("type", "notebook")
         else:
-            model.setdefault('type', 'file')
+            model.setdefault("type", "file")
 
-        insert = ''
-        if model['type'] == 'directory':
+        insert = ""
+        if model["type"] == "directory":
             untitled = self.untitled_directory
-            insert = ' '
-        elif model['type'] == 'notebook':
+            insert = " "
+        elif model["type"] == "notebook":
             untitled = self.untitled_notebook
-            ext = '.ipynb'
-        elif model['type'] == 'file':
+            ext = ".ipynb"
+        elif model["type"] == "file":
             untitled = self.untitled_file
         else:
-            raise HTTPError(400, "Unexpected model type: %r" % model['type'])
+            raise HTTPError(400, "Unexpected model type: %r" % model["type"])
 
         name = self.increment_filename(untitled + ext, path, insert=insert)
-        path = u'{0}/{1}'.format(path, name)
+        path = "{0}/{1}".format(path, name)
         return self.new(model, path)
 
-    def new(self, model=None, path=''):
+    def new(self, model=None, path=""):
         """Create a new file or directory and return its model with no content.
 
         To create a new untitled entity in a directory, use `new_untitled`.
         """
-        path = path.strip('/')
+        path = path.strip("/")
         if model is None:
             model = {}
 
-        if path.endswith('.ipynb'):
-            model.setdefault('type', 'notebook')
+        if path.endswith(".ipynb"):
+            model.setdefault("type", "notebook")
         else:
-            model.setdefault('type', 'file')
+            model.setdefault("type", "file")
 
         # no content, not a directory, so fill out new-file model
-        if 'content' not in model and model['type'] != 'directory':
-            if model['type'] == 'notebook':
-                model['content'] = new_notebook()
-                model['format'] = 'json'
+        if "content" not in model and model["type"] != "directory":
+            if model["type"] == "notebook":
+                model["content"] = new_notebook()
+                model["format"] = "json"
             else:
-                model['content'] = ''
-                model['type'] = 'file'
-                model['format'] = 'text'
+                model["content"] = ""
+                model["type"] = "file"
+                model["format"] = "text"
 
         model = self.save(model, path)
         return model
@@ -439,28 +585,28 @@ class ContentsManager(LoggingConfigurable):
 
         from_path must be a full path to a file.
         """
-        path = from_path.strip('/')
+        path = from_path.strip("/")
         if to_path is not None:
-            to_path = to_path.strip('/')
+            to_path = to_path.strip("/")
 
-        if '/' in path:
-            from_dir, from_name = path.rsplit('/', 1)
+        if "/" in path:
+            from_dir, from_name = path.rsplit("/", 1)
         else:
-            from_dir = ''
+            from_dir = ""
             from_name = path
 
         model = self.get(path)
-        model.pop('path', None)
-        model.pop('name', None)
-        if model['type'] == 'directory':
+        model.pop("path", None)
+        model.pop("name", None)
+        if model["type"] == "directory":
             raise HTTPError(400, "Can't copy directories")
 
         if to_path is None:
             to_path = from_dir
         if self.dir_exists(to_path):
-            name = copy_pat.sub(u'.', from_name)
-            to_name = self.increment_filename(name, to_path, insert='-Copy')
-            to_path = u'{0}/{1}'.format(to_path, to_name)
+            name = copy_pat.sub(".", from_name)
+            to_name = self.increment_filename(name, to_path, insert="-Copy")
+            to_path = "{0}/{1}".format(to_path, to_name)
 
         model = self.save(model, to_path)
         return model
@@ -477,12 +623,12 @@ class ContentsManager(LoggingConfigurable):
             The path of a notebook
         """
         model = self.get(path)
-        nb = model['content']
+        nb = model["content"]
         self.log.warning("Trusting notebook %s", path)
         self.notary.mark_cells(nb, True)
         self.check_and_sign(nb, path)
 
-    def check_and_sign(self, nb, path=''):
+    def check_and_sign(self, nb, path=""):
         """Check for trusted cells, and sign the notebook.
 
         Called as a part of saving notebooks.
@@ -499,7 +645,7 @@ class ContentsManager(LoggingConfigurable):
         else:
             self.log.warning("Notebook %s is not trusted", path)
 
-    def mark_trusted_cells(self, nb, path=''):
+    def mark_trusted_cells(self, nb, path=""):
         """Mark cells as trusted if the notebook signature matches.
 
         Called as a part of loading notebooks.
@@ -545,11 +691,11 @@ class AsyncContentsManager(ContentsManager):
     checkpoints = Instance(AsyncCheckpoints, config=True)
     checkpoints_kwargs = Dict(config=True)
 
-    @default('checkpoints')
+    @default("checkpoints")
     def _default_checkpoints(self):
         return self.checkpoints_class(**self.checkpoints_kwargs)
 
-    @default('checkpoints_kwargs')
+    @default("checkpoints_kwargs")
     def _default_checkpoints_kwargs(self):
         return dict(
             parent=self,
@@ -595,7 +741,7 @@ class AsyncContentsManager(ContentsManager):
         """
         raise NotImplementedError
 
-    async def file_exists(self, path=''):
+    async def file_exists(self, path=""):
         """Does a file exist at the given path?
 
         Like os.path.isfile
@@ -612,7 +758,7 @@ class AsyncContentsManager(ContentsManager):
         exists : bool
             Whether the file exists.
         """
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     async def exists(self, path):
         """Does a file or directory exist at the given path?
@@ -629,11 +775,13 @@ class AsyncContentsManager(ContentsManager):
         exists : bool
             Whether the target exists.
         """
-        return await (ensure_async(self.file_exists(path)) or ensure_async(self.dir_exists(path)))
+        return await ensure_async(self.file_exists(path)) or await ensure_async(
+            self.dir_exists(path)
+        )
 
     async def get(self, path, content=True, type=None, format=None):
         """Get a file or directory model."""
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     async def save(self, model, path):
         """
@@ -643,22 +791,22 @@ class AsyncContentsManager(ContentsManager):
         should call self.run_pre_save_hook(model=model, path=path) prior to
         writing any data.
         """
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     async def delete_file(self, path):
         """Delete the file or directory at path."""
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     async def rename_file(self, old_path, new_path):
         """Rename a file or directory."""
-        raise NotImplementedError('must be implemented in a subclass')
+        raise NotImplementedError("must be implemented in a subclass")
 
     # ContentsManager API part 2: methods that have useable default
     # implementations, but can be overridden in subclasses.
 
     async def delete(self, path):
         """Delete a file/directory and any associated checkpoints."""
-        path = path.strip('/')
+        path = path.strip("/")
         if not path:
             raise HTTPError(400, "Can't delete root")
 
@@ -676,14 +824,14 @@ class AsyncContentsManager(ContentsManager):
         For use in PATCH requests, to enable renaming a file without
         re-uploading its contents. Only used for renaming at the moment.
         """
-        path = path.strip('/')
-        new_path = model.get('path', path).strip('/')
+        path = path.strip("/")
+        new_path = model.get("path", path).strip("/")
         if path != new_path:
             await self.rename(path, new_path)
         model = await self.get(new_path, content=False)
         return model
 
-    async def increment_filename(self, filename, path='', insert=''):
+    async def increment_filename(self, filename, path="", insert=""):
         """Increment a filename until it is unique.
 
         Parameters
@@ -692,7 +840,7 @@ class AsyncContentsManager(ContentsManager):
             The name of a file, including extension
         path : unicode
             The API path of the target's directory
-        insert: unicode
+        insert : unicode
             The characters to insert after the base filename
 
         Returns
@@ -701,26 +849,27 @@ class AsyncContentsManager(ContentsManager):
             A filename that is unique, based on the input filename.
         """
         # Extract the full suffix from the filename (e.g. .tar.gz)
-        path = path.strip('/')
-        basename, dot, ext = filename.rpartition('.')
-        if ext != 'ipynb':
-                basename, dot, ext = filename.partition('.')
+        path = path.strip("/")
+        basename, dot, ext = filename.rpartition(".")
+        if ext != "ipynb":
+            basename, dot, ext = filename.partition(".")
 
         suffix = dot + ext
 
         for i in itertools.count():
             if i:
-                insert_i = '{}{}'.format(insert, i)
+                insert_i = "{}{}".format(insert, i)
             else:
-                insert_i = ''
-            name = u'{basename}{insert}{suffix}'.format(basename=basename,
-                insert=insert_i, suffix=suffix)
-            file_exists = await ensure_async(self.exists(u'{}/{}'.format(path, name)))
+                insert_i = ""
+            name = "{basename}{insert}{suffix}".format(
+                basename=basename, insert=insert_i, suffix=suffix
+            )
+            file_exists = await ensure_async(self.exists("{}/{}".format(path, name)))
             if not file_exists:
                 break
         return name
 
-    async def new_untitled(self, path='', type='', ext=''):
+    async def new_untitled(self, path="", type="", ext=""):
         """Create a new untitled file or directory in path
 
         path must be a directory
@@ -729,59 +878,59 @@ class AsyncContentsManager(ContentsManager):
 
         Use `new` to create files with a fully specified path (including filename).
         """
-        path = path.strip('/')
+        path = path.strip("/")
         dir_exists = await ensure_async(self.dir_exists(path))
         if not dir_exists:
-            raise HTTPError(404, 'No such directory: %s' % path)
+            raise HTTPError(404, "No such directory: %s" % path)
 
         model = {}
         if type:
-            model['type'] = type
+            model["type"] = type
 
-        if ext == '.ipynb':
-            model.setdefault('type', 'notebook')
+        if ext == ".ipynb":
+            model.setdefault("type", "notebook")
         else:
-            model.setdefault('type', 'file')
+            model.setdefault("type", "file")
 
-        insert = ''
-        if model['type'] == 'directory':
+        insert = ""
+        if model["type"] == "directory":
             untitled = self.untitled_directory
-            insert = ' '
-        elif model['type'] == 'notebook':
+            insert = " "
+        elif model["type"] == "notebook":
             untitled = self.untitled_notebook
-            ext = '.ipynb'
-        elif model['type'] == 'file':
+            ext = ".ipynb"
+        elif model["type"] == "file":
             untitled = self.untitled_file
         else:
-            raise HTTPError(400, "Unexpected model type: %r" % model['type'])
+            raise HTTPError(400, "Unexpected model type: %r" % model["type"])
 
         name = await self.increment_filename(untitled + ext, path, insert=insert)
-        path = u'{0}/{1}'.format(path, name)
+        path = "{0}/{1}".format(path, name)
         return await self.new(model, path)
 
-    async def new(self, model=None, path=''):
+    async def new(self, model=None, path=""):
         """Create a new file or directory and return its model with no content.
 
         To create a new untitled entity in a directory, use `new_untitled`.
         """
-        path = path.strip('/')
+        path = path.strip("/")
         if model is None:
             model = {}
 
-        if path.endswith('.ipynb'):
-            model.setdefault('type', 'notebook')
+        if path.endswith(".ipynb"):
+            model.setdefault("type", "notebook")
         else:
-            model.setdefault('type', 'file')
+            model.setdefault("type", "file")
 
         # no content, not a directory, so fill out new-file model
-        if 'content' not in model and model['type'] != 'directory':
-            if model['type'] == 'notebook':
-                model['content'] = new_notebook()
-                model['format'] = 'json'
+        if "content" not in model and model["type"] != "directory":
+            if model["type"] == "notebook":
+                model["content"] = new_notebook()
+                model["format"] = "json"
             else:
-                model['content'] = ''
-                model['type'] = 'file'
-                model['format'] = 'text'
+                model["content"] = ""
+                model["type"] = "file"
+                model["format"] = "text"
 
         model = await self.save(model, path)
         return model
@@ -796,27 +945,27 @@ class AsyncContentsManager(ContentsManager):
 
         from_path must be a full path to a file.
         """
-        path = from_path.strip('/')
+        path = from_path.strip("/")
         if to_path is not None:
-            to_path = to_path.strip('/')
+            to_path = to_path.strip("/")
 
-        if '/' in path:
-            from_dir, from_name = path.rsplit('/', 1)
+        if "/" in path:
+            from_dir, from_name = path.rsplit("/", 1)
         else:
-            from_dir = ''
+            from_dir = ""
             from_name = path
 
         model = await self.get(path)
-        model.pop('path', None)
-        model.pop('name', None)
-        if model['type'] == 'directory':
+        model.pop("path", None)
+        model.pop("name", None)
+        if model["type"] == "directory":
             raise HTTPError(400, "Can't copy directories")
         if to_path is None:
             to_path = from_dir
         if await ensure_async(self.dir_exists(to_path)):
-            name = copy_pat.sub(u'.', from_name)
-            to_name = await self.increment_filename(name, to_path, insert='-Copy')
-            to_path = u'{0}/{1}'.format(to_path, to_name)
+            name = copy_pat.sub(".", from_name)
+            to_name = await self.increment_filename(name, to_path, insert="-Copy")
+            to_path = "{0}/{1}".format(to_path, to_name)
 
         model = await self.save(model, to_path)
         return model
@@ -830,7 +979,7 @@ class AsyncContentsManager(ContentsManager):
             The path of a notebook
         """
         model = await self.get(path)
-        nb = model['content']
+        nb = model["content"]
         self.log.warning("Trusting notebook %s", path)
         self.notary.mark_cells(nb, True)
         self.check_and_sign(nb, path)
